@@ -19,9 +19,9 @@
 
 """Custom script for StarOffice and OpenOffice."""
 
-__id__        = "$Id: StarOffice.py,v 1.66 2006/12/01 00:58:58 richb Exp $"
-__version__   = "$Revision: 1.66 $"
-__date__      = "$Date: 2006/12/01 00:58:58 $"
+__id__        = "$Id: StarOffice.py 1882 2007-01-04 21:28:04Z richb $"
+__version__   = "$Revision: 1882 $"
+__date__      = "$Date: 2007-01-04 16:28:04 -0500 (Thu, 04 Jan 2007) $"
 __copyright__ = "Copyright (c) 2005-2006 Sun Microsystems Inc."
 __license__   = "LGPL"
 
@@ -50,7 +50,24 @@ inputLineForCell = None
 dynamicColumnHeaders = {}
 dynamicRowHeaders = {}
 
-def getCalcTable(obj):
+def adjustForWriterTable(obj):
+    """Check to see if we are in Writer, where the object with focus is a
+    paragraph, and the parent is the table cell. If it is, then, return the
+    parent table cell otherwise return the current object.
+
+    Arguments:
+    - obj: the accessible object to check.
+
+    Returns parent table cell (if in a Writer table ) or the current object.
+    """
+
+    if obj.role == rolenames.ROLE_PARAGRAPH and \
+       obj.parent.role == rolenames.ROLE_TABLE_CELL:
+        return obj.parent
+    else:
+        return obj
+
+def getTable(obj):
     """Get the table that this table cell is in.
 
     Arguments:
@@ -61,6 +78,7 @@ def getCalcTable(obj):
     """
 
     table = None
+    obj = adjustForWriterTable(obj)
     if obj.role == rolenames.ROLE_TABLE_CELL and obj.parent:
         table = obj.parent.table
 
@@ -77,6 +95,7 @@ def getDynamicColumnHeaderCell(obj, column):
     Return the dynamic column header cell associated with the given table cell.
     """
 
+    obj = adjustForWriterTable(obj)
     accCell = None
     parent = obj.parent
     if parent and parent.table:
@@ -97,6 +116,7 @@ def getDynamicRowHeaderCell(obj, row):
     Return the dynamic row header cell associated with the given table cell.
     """
 
+    obj = adjustForWriterTable(obj)
     accCell = None
     parent = obj.parent
     if parent and parent.table:
@@ -145,6 +165,7 @@ def isSpreadSheetCell(obj):
     Returns True if this is a table cell, False otherwise.
     """
 
+    found = False
     rolesList = [rolenames.ROLE_TABLE_CELL, \
                  rolenames.ROLE_TABLE, \
                  rolenames.ROLE_UNKNOWN, \
@@ -153,7 +174,19 @@ def isSpreadSheetCell(obj):
                  rolenames.ROLE_ROOT_PANE, \
                  rolenames.ROLE_FRAME, \
                  rolenames.ROLE_APPLICATION]
-    return util.isDesiredFocusedItem(obj, rolesList)
+    if util.isDesiredFocusedItem(obj, rolesList):
+        # We've found a table cell with the correct hierarchy. Now check
+        # that we are in a spreadsheet as opposed to the writer application.
+        # See bug #382408.
+        #
+        current = obj.parent
+        while current.role != rolenames.ROLE_APPLICATION:
+            if current.role == rolenames.ROLE_FRAME and \
+               (current.name and current.name.endswith(_("Calc"))):
+                found = True
+            current = current.parent
+
+    return found
 
 class BrailleGenerator(braillegenerator.BrailleGenerator):
     """Overrides _getBrailleRegionsForTableCellRow so that , when we are
@@ -183,36 +216,48 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
 
         regions = []
 
+        # Check to see if this spread sheet cell has either a dynamic
+        # column heading or row heading (or both) associated with it.
+        # If it does, then braille those first before brailling the
+        # cell contents.
+        #
+        table = getTable(obj)
+        parent = obj.parent
+
+        if parent.__dict__.has_key("lastColumn") and \
+           parent.lastColumn != parent.table.getColumnAtIndex(obj.index):
+            if dynamicColumnHeaders.has_key(table):
+                row = dynamicColumnHeaders[table]
+                header = getDynamicRowHeaderCell(obj, row)
+                if header.childCount > 0:
+                    for i in range(0, header.childCount):
+                        child = header.child(i)
+                        text = self._script.getText(child, 0, -1)
+                        if text:
+                            regions.append(braille.Region(" " + text + " "))
+                elif header.text:
+                    text = self._script.getText(header, 0, -1)
+                    if text:
+                        regions.append(braille.Region(" " + text + " "))
+
+        if parent.__dict__.has_key("lastRow") and \
+           parent.lastRow != parent.table.getRowAtIndex(obj.index):
+            if dynamicRowHeaders.has_key(table):
+                column = dynamicRowHeaders[table]
+                header = getDynamicColumnHeaderCell(obj, column)
+                if header.childCount > 0:
+                    for i in range(0, header.childCount):
+                        child = header.child(i)
+                        text = self._script.getText(child, 0, -1)
+                        if text:
+                            regions.append(braille.Region(" " + text + " "))
+                elif header.text:
+                    text = self._script.getText(header, 0, -1)
+                    if text:
+                        regions.append(braille.Region(" " + text + " "))
+
         if isSpreadSheetCell(obj):
 
-            # Check to see if this spread sheet cell has either a dynamic
-            # column heading or row heading (or both) associated with it.
-            # If it does, then braille those first before brailling the
-            # cell contents.
-            #
-            table = getCalcTable(obj)
-            parent = obj.parent
-
-            if parent.__dict__.has_key("lastColumn") and \
-               parent.lastColumn != parent.table.getColumnAtIndex(obj.index):
-                if dynamicColumnHeaders.has_key(table):
-                    row = dynamicColumnHeaders[table]
-                    header = getDynamicRowHeaderCell(obj, row)
-                    if header.text:
-                        text = self._script.getText(header, 0, -1)
-                        if text:
-                            regions.append(braille.Region(" " + text + " "))
-
-            if parent.__dict__.has_key("lastRow") and \
-               parent.lastRow != parent.table.getRowAtIndex(obj.index):
-                if dynamicRowHeaders.has_key(table):
-                    column = dynamicRowHeaders[table]
-                    header = getDynamicColumnHeaderCell(obj, column)
-                    if header.text:
-                        text = self._script.getText(header, 0, -1)
-                        if text:
-                            regions.append(braille.Region(" " + text + " "))
-
             # Adding in a check here to make sure that the parent is a
             # valid table. It's possible that the parent could be a
             # table cell too (see bug #351501).
@@ -267,7 +312,10 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
             regions = [regions, focusRegion]
         else:
             brailleGen = braillegenerator.BrailleGenerator
-            regions = brailleGen._getBrailleRegionsForTableCellRow(self, obj)
+            [cellRegions, focusRegion] = \
+                brailleGen._getBrailleRegionsForTableCellRow(self, obj)
+            regions.extend(cellRegions)
+            regions = [regions, focusRegion]
 
         return regions
 
@@ -307,8 +355,24 @@ class BrailleGenerator(braillegenerator.BrailleGenerator):
             return [regions, componentRegion]
 
         else:
-            brailleGen = braillegenerator.BrailleGenerator
-            regions = brailleGen._getBrailleRegionsForTableCell(self, obj)
+            # Check to see how many children this table cell has. If it's
+            # just one (or none), then pass it on to the superclass to be
+            # processed.
+            #
+            # If it's more than one, then get the braille regions for each 
+            # child, and call this method again.
+            #
+            if obj.childCount <= 1:
+                brailleGen = braillegenerator.BrailleGenerator
+                regions = brailleGen._getBrailleRegionsForTableCell(self, obj)
+            else:
+                regions = []
+                for i in range(0, obj.childCount):
+                    child = obj.child(i)
+                    [cellRegions, focusRegion] = \
+                                self._getBrailleRegionsForTableCell(child)
+                    regions.extend(cellRegions)
+                return [regions, focusRegion]
 
         return regions
 
@@ -382,38 +446,51 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
 
         utterances = []
 
+        if not already_focused:
+
+            # Check to see if this spread sheet cell has either a dynamic
+            # column heading or row heading (or both) associated with it.
+            # If it does, then speak those first before speaking the cell
+            # contents.
+            #
+            table = getTable(obj)
+            parent = obj.parent
+
+            if parent.__dict__.has_key("lastColumn") and \
+               parent.lastColumn != \
+               parent.table.getColumnAtIndex(obj.index):
+                if dynamicColumnHeaders.has_key(table):
+                    row = dynamicColumnHeaders[table]
+                    header = getDynamicRowHeaderCell(obj, row)
+                    if header.childCount > 0:
+                        for i in range(0, header.childCount):
+                            child = header.child(i)
+                            text = self._script.getText(child, 0, -1)
+                            if text:
+                                utterances.append(text)
+                    elif header.text:
+                        text = self._script.getText(header, 0, -1)
+                        if text:
+                            utterances.append(text)
+
+            if parent.__dict__.has_key("lastRow") and \
+               parent.lastRow != parent.table.getRowAtIndex(obj.index):
+                if dynamicRowHeaders.has_key(table):
+                    column = dynamicRowHeaders[table]
+                    header = getDynamicColumnHeaderCell(obj, column)
+                    if header.childCount > 0:
+                        for i in range(0, header.childCount):
+                            child = header.child(i)
+                            text = self._script.getText(child, 0, -1)
+                            if text:
+                                utterances.append(text)
+                    elif header.text:
+                        text = self._script.getText(header, 0, -1)
+                        if text:
+                            utterances.append(text)
+
         if isSpreadSheetCell(obj):
-            if (not already_focused):
-
-                # Check to see if this spread sheet cell has either a dynamic
-                # column heading or row heading (or both) associated with it.
-                # If it does, then speak those first before speaking the cell
-                # contents.
-                #
-                table = getCalcTable(obj)
-                parent = obj.parent
-
-                if parent.__dict__.has_key("lastColumn") and \
-                   parent.lastColumn != \
-                   parent.table.getColumnAtIndex(obj.index):
-                    if dynamicColumnHeaders.has_key(table):
-                        row = dynamicColumnHeaders[table]
-                        header = getDynamicRowHeaderCell(obj, row)
-                        if header.text:
-                            text = self._script.getText(header, 0, -1)
-                            if text:
-                                utterances.append(text)
-
-                if parent.__dict__.has_key("lastRow") and \
-                   parent.lastRow != parent.table.getRowAtIndex(obj.index):
-                    if dynamicRowHeaders.has_key(table):
-                        column = dynamicRowHeaders[table]
-                        header = getDynamicColumnHeaderCell(obj, column)
-                        if header.text:
-                            text = self._script.getText(header, 0, -1)
-                            if text:
-                                utterances.append(text)
-
+            if not already_focused:
                 if settings.readTableCellRow:
                     parent = obj.parent
                     row = parent.table.getRowAtIndex(obj.index)
@@ -448,8 +525,8 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
                                                              already_focused))
         else:
             speechGen = speechgenerator.SpeechGenerator
-            utterances = speechGen._getSpeechForTableCellRow(self, obj,
-                                                             already_focused)
+            utterances.extend(speechGen._getSpeechForTableCellRow(self, obj,
+                                                             already_focused))
 
         return utterances
 
@@ -480,9 +557,23 @@ class SpeechGenerator(speechgenerator.SpeechGenerator):
             nameList = obj.name.split()
             utterances.append(nameList[1])
         else:
-            speechGen = speechgenerator.SpeechGenerator
-            utterances = speechGen._getSpeechForTableCell(self, obj,
-                                                      already_focused)
+            # Check to see how many children this table cell has. If it's
+            # just one (or none), then pass it on to the superclass to be 
+            # processed.
+            #
+            # If it's more than one, then get the speech for each child,
+            # and call this method again.
+            #
+            if obj.childCount <= 1:
+                speechGen = speechgenerator.SpeechGenerator
+                utterances = speechGen._getSpeechForTableCell(self, obj,
+                                                        already_focused)
+            else:
+                utterances = []
+                for i in range(0, obj.childCount):
+                    child = obj.child(i)
+                    utterances.extend(self._getSpeechForTableCell(child,
+                                                        already_focused))
 
         return utterances
 
@@ -529,6 +620,11 @@ class Script(default.Script):
         #
         self.currentParagraph = None
 
+        # The default set of text attributes to speak to the user. The
+        # only difference over the default set in settings.py is to add
+        # in "left-margin:" and "right-margin:".
+
+        settings.enabledTextAttributes = "size:; family-name:; weight:400; indent:0; left-margin:0; right-margin:0; underline:none; strikethrough:false; justification:left; style:normal;"
 
     def getBrailleGenerator(self):
         """Returns the braille generator for this script.
@@ -598,6 +694,41 @@ class Script(default.Script):
 
         return keyBindings
 
+    def checkForTableBoundry(self, oldFocus, newFocus):
+        """Check to see if we've entered or left a table.
+        When entering a table, announce the table dimensions.
+        When leaving a table, announce that the table has been exited.
+
+        Arguments:
+        - oldFocus: Accessible that is the old locus of focus
+        - newFocus: Accessible that is the new locus of focus
+        """
+
+        if oldFocus == None or newFocus == None:
+            return
+
+        oldFocusIsTable = None 
+        while oldFocus.role != rolenames.ROLE_APPLICATION:
+            if oldFocus.role == rolenames.ROLE_TABLE:
+                oldFocusIsTable = oldFocus
+                break
+            oldFocus = oldFocus.parent 
+
+        newFocusIsTable = None
+        while newFocus.role != rolenames.ROLE_APPLICATION:
+            if newFocus.role == rolenames.ROLE_TABLE:
+                newFocusIsTable = newFocus
+                break
+            newFocus = newFocus.parent
+
+        if oldFocusIsTable == None and newFocusIsTable != None:
+            rows = newFocusIsTable.table.nRows
+            columns = newFocusIsTable.table.nColumns
+            line = _("table with %d rows and %d columns.") % (rows, columns)
+            speech.speak(line)
+        elif oldFocusIsTable != None and newFocusIsTable == None:
+            speech.speak(_("leaving table."))
+
     def speakInputLine(self, inputEvent):
         """Speak the contents of the spread sheet input line (assuming we
         have a handle to it - generated when we first focus on a spread
@@ -623,7 +754,7 @@ class Script(default.Script):
                         "StarOffice.speakInputLine: contents: %s" % inputLine)
                 speech.speak(inputLine)
 
-    def getCalcRow(self, cell):
+    def getTableRow(self, cell):
         """Get the row number in the table that this table cell is on.
 
         Arguments:
@@ -634,6 +765,7 @@ class Script(default.Script):
         """
 
         row = None
+        cell = adjustForWriterTable(cell)
         if cell.role == rolenames.ROLE_TABLE_CELL:
             parent = cell.parent
             if parent and parent.table:
@@ -641,7 +773,7 @@ class Script(default.Script):
 
         return row
 
-    def getCalcColumn(self, cell):
+    def getTableColumn(self, cell):
         """Get the column number in the table that this table cell is on.
 
         Arguments:
@@ -652,6 +784,7 @@ class Script(default.Script):
         """
 
         column = None
+        cell = adjustForWriterTable(cell)
         if cell.role == rolenames.ROLE_TABLE_CELL:
             parent = cell.parent
             if parent and parent.table:
@@ -674,14 +807,14 @@ class Script(default.Script):
         - inputEvent: if not None, the input event that caused this action.
         """
 
-        global dynamicColumnHeaders, getCalcTable
+        global dynamicColumnHeaders, getTable
 
         debug.println(self.debugLevel, "StarOffice.setDynamicColumnHeaders.")
 
         clickCount = util.getClickCount(self.lastDynamicEvent, inputEvent)
-        table = getCalcTable(orca_state.locusOfFocus)
+        table = getTable(orca_state.locusOfFocus)
         if table:
-            row = self.getCalcRow(orca_state.locusOfFocus)
+            row = self.getTableRow(orca_state.locusOfFocus)
             if clickCount == 2:
                 try:
                     del dynamicColumnHeaders[table]
@@ -737,14 +870,14 @@ class Script(default.Script):
         - inputEvent: if not None, the input event that caused this action.
         """
 
-        global dynamicRowHeaders, getCalcTable
+        global dynamicRowHeaders, getTable
 
         debug.println(self.debugLevel, "StarOffice.setDynamicRowHeaders.")
 
         clickCount = util.getClickCount(self.lastDynamicEvent, inputEvent)
-        table = getCalcTable(orca_state.locusOfFocus)
+        table = getTable(orca_state.locusOfFocus)
         if table:
-            column = self.getCalcColumn(orca_state.locusOfFocus)
+            column = self.getTableColumn(orca_state.locusOfFocus)
             if clickCount == 2:
                 try:
                     del dynamicRowHeaders[table]
@@ -1305,6 +1438,18 @@ class Script(default.Script):
             orca.setLocusOfFocus(None, event.source.parent, False)
             return
 
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then speak/braille that parent table cell (see bug #382415).
+        #
+        if event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+            if self.lastCell != event.source.parent:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source.parent)
+                self.lastCell = event.source.parent
+            return
+
         default.Script.onFocus(self, event)
 
     def onStateChanged(self, event):
@@ -1314,6 +1459,19 @@ class Script(default.Script):
         - event: the Event
         """
 
+        # If this is a state change "focused" event that we care about, and 
+        # we are in Writer, check to see if we are entering or leaving a table.
+        #
+        if event.type == "object:state-changed:focused" and event.detail1 == 1:
+            current = event.source.parent
+            while current.role != rolenames.ROLE_APPLICATION:
+                if current.role == rolenames.ROLE_FRAME and \
+                   (current.name and current.name.endswith(_("Writer"))):
+                    self.checkForTableBoundry(orca_state.locusOfFocus, 
+                                              event.source)
+                    break
+                current = current.parent
+
         # Prevent  "object:state-changed:active" events from activating
         # the find operation. See comment #18 of bug #354463.
         #
@@ -1321,6 +1479,35 @@ class Script(default.Script):
             if self.findCommandRun:
                 return
 
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then speak/braille that parent table cell (see bug #382415).
+        #
+        if event.type == "object:state-changed:focused" and \
+           event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.detail1 == 0 and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+
+            # Check to see if the last input event was "Up" or "Down".
+            # If it was, and we are in the same table cell as last time,
+            # and if that table cell has more than one child, then just
+            # get the speech for that single child, otherwise speak/braille
+            # the parent table cell.
+            #
+            event_string = None
+            if isinstance(orca_state.lastInputEvent, input_event.KeyboardEvent):
+                event_string = orca_state.lastInputEvent.event_string
+            if (event_string == "Up" or event_string == "Down") and \
+               event.source.parent == self.lastCell and \
+               event.source.parent.childCount > 1:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source)
+            else:
+                default.Script.locusOfFocusChanged(self, event,
+                                                   None, event.source.parent)
+            self.lastCell = event.source.parent
+            return
+
         # Two events are received when the caret moves
         # to a new paragraph. The first is a focus event
         # (in the form of object:state-changed:focused
@@ -1339,7 +1526,8 @@ class Script(default.Script):
         # to help reduce the verbosity of focusing on the Calc Name combo 
         # box (see bug #364407).
         #
-        if event.source.parent.role == rolenames.ROLE_COMBO_BOX:
+        if event.source.parent and \
+           event.source.parent.role == rolenames.ROLE_COMBO_BOX:
             orca.setLocusOfFocus(None, event.source.parent, False)
             return
 
@@ -1437,6 +1625,16 @@ class Script(default.Script):
         string = string.encode("UTF-8")
         return string
 
+    def speakCellName(self, name):
+        """Speaks the given cell name.
+
+        Arguments:
+        - name: the name of the cell
+        """
+
+        line = _("Cell ") + name
+        speech.speak(line)
+
     def onCaretMoved(self, event):
         """Called whenever the caret moves.
 
@@ -1444,6 +1642,63 @@ class Script(default.Script):
         - event: the Event
         """
 
+        # If we are FOCUSED on a paragraph inside a table cell (in Writer),
+        # then just return (modulo the special cases below). Speaking and 
+        # brailling will have been done in the onStateChanged() routine 
+        # (see bug #382415).
+        #
+        if event.source.role == rolenames.ROLE_PARAGRAPH and \
+           event.source.parent.role == rolenames.ROLE_TABLE_CELL and \
+           event.source.state.count(atspi.Accessibility.STATE_FOCUSED):
+            event_string = orca_state.lastInputEvent.event_string
+
+            # If we are moving up and down, and we are speaking-by-cell
+            # (as opposed to by-row), then speak the cell name. Otherwise
+            # just return.
+            #
+            if (event_string == "Up" or event_string == "Down"):
+                if settings.readTableCellRow == False:
+                    if event.detail1 != -1:
+                        self.speakCellName(event.source.parent.name)
+                return
+
+            # If we are moving left or right and we are in a new cell, just
+            # return.
+            #
+            if (event_string == "Left" or event_string == "Right") and \
+               self.lastCell != event.source.parent:
+                return
+
+            caretOffset = event.source.text.caretOffset
+            len = event.source.text.characterCount
+
+            # If you are in a table cell and you arrow Right, the caret 
+            # will focus at the end of the current paragraph before moving 
+            # into the next cell. To be similar to the way that caret 
+            # navigation works in other paragraphs in OOo, just return.
+            #
+            if event_string == "Right" and caretOffset == len:
+                return
+
+            # If we have moved left and the caret position is at the end of
+            # the paragraph or if we have moved right and the caret position
+            # is at the start of the text string, or the last key input was
+            # Tab or Shift-Tab, and if we are speaking-by-cell (as opposed 
+            # to by-row), then speak the cell name, otherwise just return 
+            # (see bug #382418).
+            #
+            if (event_string == "Left" and caretOffset == len) or \
+               (event_string == "Right" and caretOffset == 0) or \
+               (event_string == "Tab" or event_string == "ISO_Left_Tab"):
+                if settings.readTableCellRow == False:
+                    if event.detail1 != -1:
+                        self.speakCellName(event.source.parent.name)
+
+                # Speak a blank line, if appropriate.
+                if self.speakBlankLine(event.source):
+                    speech.speak(_("blank"), None, False)
+                return
+
         # Speak a newline, if appropriate.
         if self.speakNewLine(event.source):
             speech.speak(chnames.getCharacterName("\n"), None, False)
